import typing as t
import time
import hmac
from .core import OTP, SupportedAlgorithms
from .rfc4226 import generate_hotp


class TOTP(OTP):
    """Implementation of RFC6238, Time-Based One-Time Password Algorithm.

    :param secret: A secret in bytes
    :param digit: Number of digits in the HOTP code.
    :param algorithm: Hash algorithm used in HOTP.
    :param period: The password valid in "period" seconds.
    """

    TYPE: t.ClassVar[str] = "TOTP"

    def __init__(self, secret: bytes, digit: int = 6, algorithm: SupportedAlgorithms = "SHA1", period: int = 30):
        super().__init__(secret, digit, algorithm)
        self.period = period

    def generate(self, timestamp: t.Optional[int] = None) -> int:
        """Generate a TOTP code. The returning result is an integer.
        To convert it into string with the correct digit length, developers
        can use :meth:`string_code`::

            int_code = totp.generate()
            str_code = totp.string_code(int_code)

        :param timestamp: Create TOTP at this given timestamp, default is now.
        """
        return generate_totp(self.secret, self.period, timestamp, self.digit, self.algorithm)

    def verify(self, code: int, timestamp: t.Optional[int] = None) -> bool:  # type: ignore[override]
        """Valid a TOTP code for the given timestamp.

        :param code: A number to be verified.
        :param timestamp: Validate TOTP at this given timestamp, default is now.
        """
        if len(str(code)) > self.digit:
            return False
        return hmac.compare_digest(self.string_code(self.generate(timestamp)), self.string_code(code))

    def to_uri(self, label: str, issuer: str) -> str:  # type: ignore[override]
        """Generate the otpauth protocal string for TOTP.

        :param label: Label of the identifier.
        :param issuer: The company, the organization or something else.
        """
        uri = self._get_base_uri(label, issuer)
        return uri + f"&period={self.period}"


def generate_totp(
        secret: bytes,
        period: int = 30,
        timestamp: t.Optional[int] = None,
        digit: int = 6,
        algorithm: SupportedAlgorithms = "SHA1") -> int:
    """Generate a TOTP code.

    A TOTP code is an extension of TOTP algorithm.

    :param secret: A secret token for the authentication.
    :param period: A period that a TOTP code is valid in seconds
    :param timestamp: Create TOTP at this given timestamp, default is now.
    """
    if timestamp is None:
        timestamp = int(time.time())
    counter = int(timestamp) // period
    return generate_hotp(secret, counter, digit, algorithm)



import unittest
from otpauth import HOTP, TOTP


class TestMisc(unittest.TestCase):
    def test_invalid_digit(self):
        self.assertRaises(AssertionError, HOTP, b"secret", 0)
        self.assertRaises(AssertionError, TOTP, b"secret", 0)

        self.assertRaises(AssertionError, HOTP, b"secret", 11)
        self.assertRaises(AssertionError, TOTP, b"secret", 11)

    def test_invalid_algorithm(self):
        self.assertRaises(AssertionError, HOTP, b"secret", algorithm="SHA2")
        self.assertRaises(AssertionError, TOTP, b"secret", algorithm="SHA2")

    def test_string_code(self):
        htop = HOTP(b"secret")
        self.assertEqual(htop.string_code(123), "000123")



import base64
import typing as t
from urllib.parse import quote
from abc import ABCMeta, abstractmethod

SupportedAlgorithms = t.Literal["SHA1", "SHA256", "SHA512"]
Self = t.TypeVar("Self", bound="OTP")


class OTP(metaclass=ABCMeta):
    TYPE: t.ClassVar[str]

    #: The supportted algorithms
    ALGORITHMS: t.ClassVar[t.List[str]] = ["SHA1", "SHA256", "SHA512"]

    def __init__(self, secret: bytes, digit: int = 6, algorithm: SupportedAlgorithms = "SHA1"):
        assert 0 < digit < 11
        assert algorithm in self.ALGORITHMS

        self.secret = secret
        self.digit = digit
        self.algorithm = algorithm
        self._b32_secret: t.Optional[str] = None

    @property
    def b32_secret(self) -> str:
        if self._b32_secret:
            return self._b32_secret

        secret = base64.b32encode(self.secret)
        self._b32_secret = secret.rstrip(b"=").decode("ascii")
        return self._b32_secret

    @classmethod
    def from_b32encode(
            cls: t.Type[Self],
            secret: t.Union[bytes, str],
            digit: int = 6,
            algorithm: SupportedAlgorithms = "SHA1") -> Self:
        """Create the instance with a base32 encoded secret.

        :param secret: A base32 encoded secret string or bytes.
        :param digit: Number of digits in the OTP code.
        :param algorithm: Hash algorithm used in HOTP.
        """
        if isinstance(secret, str):
            secret = secret.encode("utf-8")

        b32_secret = secret.rstrip(b"=")

        # add padding back
        secret += b"=" * (-len(secret) % 8)
        raw_secret = base64.b32decode(secret)

        obj = cls(raw_secret, digit, algorithm)
        obj._b32_secret = b32_secret.decode("ascii")
        return obj

    def _get_base_uri(self, label: str, issuer: str) -> str:
        label = quote(label, safe="/@:")
        issuer = quote(issuer, safe="")
        _type = self.TYPE.lower()
        url = (
            f"otpauth://{_type}/{label}"
            f"?secret={self.b32_secret}"
            f"&issuer={issuer}"
            f"&algorithm={self.algorithm}"
            f"&digits={self.digit}"
        )
        return url

    def string_code(self, code: int) -> str:
        """Add leading 0 if the code length does not match the defined length.

        For instance, parameter ``digit=6``, but ``code=123``, this method would
        return ``000123``::

            >>> otp.string_code(123)
            '000123'

        :param code: The number that this OTP generated.
        """
        return f'{code:0{self.digit}}'

    @abstractmethod
    def generate(self, *args: t.Any, **kwargs: t.Any) -> int:
        ...

    @abstractmethod
    def verify(self, code: int, *args: t.Any, **kwargs: t.Any) -> bool:
        ...

    @abstractmethod
    def to_uri(self, label: str, issuer: str, *args: t.Any, **kwargs: t.Any) -> str:
        ...



import unittest
from otpauth import HOTP


class TestHOTP(unittest.TestCase):
    def setUp(self):
        self.hotp = HOTP(b"python")

    def test_generate(self):
        value = self.hotp.generate(0)
        self.assertEqual(value, 170566)

    def test_verify(self):
        # due to number too long
        self.assertFalse(self.hotp.verify(12345678, 0))

        # due to not match
        self.assertFalse(self.hotp.verify(12345, 0))

        self.assertTrue(self.hotp.verify(170566, 0))

    def test_to_uri(self):
        uri = self.hotp.to_uri("Typlog:lepture.com", "Authlib", 0)
        expected = "otpauth://hotp/Typlog:lepture.com?secret=OB4XI2DPNY&issuer=Authlib&algorithm=SHA1&digits=6&counter=0"
        self.assertEqual(uri, expected)

    def test_from_b32encode(self):
        expected = "otpauth://hotp/Typlog:lepture.com?secret=OB4XI2DPNY&issuer=Authlib&algorithm=SHA1&digits=6&counter=0"

        hotp = HOTP.from_b32encode("OB4XI2DPNY")
        value = hotp.generate(0)
        self.assertEqual(value, 170566)
        uri = self.hotp.to_uri("Typlog:lepture.com", "Authlib", 0)
        self.assertEqual(uri, expected)

        hotp = HOTP.from_b32encode("OB4XI2DPNY======")
        value = hotp.generate(0)
        self.assertEqual(value, 170566)
        uri = self.hotp.to_uri("Typlog:lepture.com", "Authlib", 0)
        self.assertEqual(uri, expected)

        hotp = HOTP.from_b32encode(b"OB4XI2DPNY======")
        value = hotp.generate(0)
        self.assertEqual(value, 170566)
        uri = self.hotp.to_uri("Typlog:lepture.com", "Authlib", 0)
        self.assertEqual(uri, expected)



import typing as t
import struct
import hmac
import hashlib
from .core import OTP, SupportedAlgorithms


class HOTP(OTP):
    """Implementation of RFC4226, An HMAC-Based One-Time Password Algorithm.

    :param secret: A secret in bytes
    :param digit: Number of digits in the HOTP code.
    :param algorithm: Hash algorithm used in HOTP.
    """

    TYPE: t.ClassVar[str] = "HOTP"

    def generate(self, counter: int) -> int:
        """Generate a HOTP code. The returning result is an integer.
        To convert it into string with the correct digit length, developers
        can use :meth:`string_code`::

            int_code = hotp.generate(4)
            str_code = hotp.string_code(int_code)

        :param counter: HOTP is a counter based algorithm.
        """
        return generate_hotp(self.secret, counter, self.digit, self.algorithm)

    def verify(self, code: int, counter: int) -> bool:  # type: ignore[override]
        """Valid a HOTP code at the given counter.

        :param code: A number to be verified.
        :param counter: The counter HOTP code based on.
        """
        if len(str(code)) > self.digit:
            return False
        return hmac.compare_digest(self.string_code(self.generate(counter)), self.string_code(code))

    def to_uri(self, label: str, issuer: str, counter: int) -> str:  # type: ignore[override]
        """Generate the otpauth protocal string for HOTP.

        :param label: Label of the identifier.
        :param issuer: The company, the organization or something else.
        :param counter: Initial counter of the HOTP algorithm.
        """
        uri = self._get_base_uri(label, issuer)
        return uri + f"&counter={counter}"


def generate_hotp(secret: bytes, counter: int, digit: int = 6, algorithm: SupportedAlgorithms = "SHA1") -> int:
    """Generate a HOTP code.

    :param secret: A secret token for the authentication.
    :param counter: HOTP is a counter based algorithm.
    :param digit: Number of digits in the HOTP code.
    :param algorithm: Hash algorithm used in HOTP.
    """
    hash_alg = getattr(hashlib, algorithm.lower())
    msg = struct.pack(">Q", counter)
    digest = hmac.new(secret, msg, hash_alg).digest()
    offset = digest[19] & 0xF
    bin_code: int = struct.unpack(">I", digest[offset: offset + 4])[0]
    total: int = bin_code & 0x7FFFFFFF
    power: int = 10 ** digit
    return total % power



import unittest
import time
from otpauth import TOTP

FIXED_TIME = 1679576495


class TestTOTP(unittest.TestCase):
    def setUp(self):
        self.totp = TOTP(b"python")

    def test_generate(self):
        value = self.totp.generate(FIXED_TIME)
        self.assertEqual(value, 129815)

    def test_verify(self):
        # due to number too long
        self.assertFalse(self.totp.verify(12345678, FIXED_TIME))

        # due to not match
        self.assertFalse(self.totp.verify(12345, FIXED_TIME))

        self.assertTrue(self.totp.verify(129815, FIXED_TIME))

    def test_to_uri(self):
        uri = self.totp.to_uri("Typlog:lepture.com", "Authlib")
        expected = "otpauth://totp/Typlog:lepture.com?secret=OB4XI2DPNY&issuer=Authlib&algorithm=SHA1&digits=6&period=30"
        self.assertEqual(uri, expected)

    def test_current_timestamp(self):
        value = self.totp.generate()
        self.assertTrue(self.totp.verify(value))



from .core import SupportedAlgorithms
from .rfc4226 import HOTP, generate_hotp
from .rfc6238 import TOTP, generate_totp


__author__ = "Hsiaoming Yang <me@lepture.com>"
__homepage__ = "https://github.com/lepture/otpauth"
__version__ = "2.1.1"

__all__ = [
    "SupportedAlgorithms",
    "HOTP",
    "TOTP",
    "generate_hotp",
    "generate_totp",
]
